import numpy as np
from IPython import display
import ipywidgets as widgets
import pythreejs
import math
import time
import matplotlib
import matplotlib.pyplot as plt
import io

desactiver_3d = False

# outils
def radians(deg):
    return deg / 180 * math.pi

def degres(rad):
    return rad / math.pi * 180

def polaires(pos, centre):
    direction = np.array(pos) - np.array(centre)
    distance = np.linalg.norm(direction)
    direction /= distance
    phi = math.asin(direction[1])
    costheta = max(-1.,min(1.,direction[0] / math.cos(phi)))
    sintheta = max(-1.,min(1.,direction[2] / math.cos(phi)))
    theta = math.acos(costheta)
    if sintheta < 0:
        theta = 2*math.pi - theta
    return (theta, phi, distance)  

def cartesiennes(theta, phi, distance, centre):
    x = centre[0] + distance * math.cos(phi) * math.cos(theta)
    y = centre[1] + distance * math.sin(phi)
    z = centre[2] + distance * math.cos(phi) * math.sin(theta)
    return (x,y,z)

# boîte englobante alignée sur les axes
class Box:
    def __init__(self, point):
        #la boîte est représentée par deux de ses points diagonalement opposés
        self.min = np.array(point)
        self.max = np.array(point)
        #nous lui ajoutons une couleur pour la dessiner
        self.color = np.array((0.2, 0.2, 0.2))
        
    def push(self, point):
        #la boîte va grossir pour englober le point
        self.min = np.minimum(self.min, np.array(point))
        self.max = np.maximum(self.max, np.array(point))
        
    def raytrace(self, origin, direction):
        #intersection rayon boîte
        inf = None
        infid = None
        sup = None
        supid = None

        #intervalles d'intersection
        for c in range(3):
            if direction[c] == 0:
                if origin[c] < self.min[c] or origin[c] > self.max[c]:
                    return None, None
            else:
                minhit = (self.min[c] - origin[c]) / direction[c]
                maxhit = (self.max[c] - origin[c]) / direction[c]
                cinf = min(minhit, maxhit)
                csup = max(minhit, maxhit)
                if inf is None or cinf > inf:
                    infid = c
                    inf = cinf
                if sup is None or csup < sup:
                    supid = c
                    sup = csup
        
        if inf < sup:
            if sup > 0:
                if inf > 0:
                    return inf,infid
                return sup,supid
        return None, None
    
    def mesh(self):
        if desactiver_3d:
            return None
        dim = self.max - self.min
        center = 0.5*(self.min + self.max)
        rgbcolor = (self.color * 255).astype(int)
        return pythreejs.Mesh(
            pythreejs.BoxGeometry(*dim), #taille de la boite (largeur, hauteur, profondeur)
            pythreejs.MeshPhysicalMaterial(color=f"rgb{tuple(rgbcolor)}", roughness = 1), #couleur de la boite
            position = tuple(center) #position 3D du centre de la boite
        )
    
# caméra trou d'épingle
class Camera:
    def __init__(self):
        object.__setattr__(self,'_observers', [])
        object.__setattr__(self,'position', np.array((7.,7.,7.)))
        object.__setattr__(self,'centre', np.array((0.,1.,0.)))
        object.__setattr__(self,'up', np.array((0.,1.,0.)))
        object.__setattr__(self,'fov', 50)
        object.__setattr__(self,'height', 600)
        object.__setattr__(self,'width', 600)
        
        self._VI = None
        self._V = None
        self._P = None
        self._PI = None
        self._updateMatrices()

        self._locked = False

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if name[0] != '_':
            self._updateMatrices()
            for o in self._observers:
                self._sync(o)
            
        
    def _updateMatrices(self):
        self._VI = self._viewInverse()
        self._V = np.linalg.inv(self._VI)
        self._P = self._projection()
        self._PI = np.linalg.inv(self._P)
        
    def _viewInverse(self):
        direction = (self.centre - self.position)
        direction /= np.linalg.norm(direction)
        right = np.cross(direction, self.up)
        right /= np.linalg.norm(right)
        newUp = np.cross(right, direction)
        newUp /= np.linalg.norm(newUp)

        VI = np.array([
            [right[0], newUp[0], -direction[0], self.position[0]],
            [right[1], newUp[1], -direction[1], self.position[1]],
            [right[2], newUp[2], -direction[2], self.position[2]],
            [0,       0,        0,     1]
        ])
        return VI
    
    def _projection(self):
        itan= 1/math.tan(radians(self.fov) * 0.5);
        ratio = self.width / self.height
        P = np.array([
            [itan/ratio, 0, 0, 0],
            [0, itan, 0, 0],
            [0, 0, -3, -4],
            [0, 0, -1,  0]
        ])
        return P
    
    def point(self, pixel):
        px = np.array([
            2 * pixel[0] / self.width - 1,
            -2 * pixel[1] / self.width + 1,
            1, 
            1
        ])
        pt = self._VI@self._PI@px
        return pt[:3] / pt[3]
    
    def pixel(self, point):
        px = self._P@self._V@np.array([*point, 1])
        px /= px[3]
        px += 1
        px *= np.array([self.width/2, self.height/2, 1,0])
        return np.array([px[0], self.height - px[1]], dtype='int')

    def sync(self, cam):
        self._observers.append(cam)
        self._sync(cam)
    
    def _sync(self,cam):
        cam.fov = self.fov
        cam.position = tuple(self.position)
        cam.lookAt(tuple(self.centre))

    def dessiner(self, scene):
        img = np.zeros((self.width,self.height,3), dtype='float')
        light_direction = np.array((1.,1.,0))
        light_direction /= np.linalg.norm(light_direction)

        for x in range(self.width):
            for y in range(self.height):
                pt = self.point((x,y))
                direction = pt - self.position
                direction /= np.linalg.norm(direction)
                hit, index, obj = None, None, None
                for b in scene:
                    h,i = b.raytrace(self.position, direction)
                    if h:
                        if hit is None or h < hit:
                            hit = h
                            index = i
                            obj = b
                if hit is None:
                    img[y,x,:] = (1,1,1)
                else:
                    normal = np.zeros(3)
                    normal[index] = 1 if direction[index] < 0 else -1
                    diffuse = max(0, light_direction.dot(normal))
                    
                    hv = light_direction - direction
                    hv /= np.linalg.norm(direction)
                    specular = max(0, hv.dot(normal))

                    coeff = 1*diffuse + 0.2*specular + 0.8
                    img[y,x,:] = np.minimum(obj.color * coeff, np.ones(3))

        return img

    def application(self, scene, souris = False):
        if desactiver_3d:
            return None
        meshes = [o.mesh() for o in scene]
        
        lumiere_ambiante = pythreejs.AmbientLight(intensity = 1.5) #lumiere globale uniforme
        soleil = pythreejs.DirectionalLight(position=[10,10, 0], intensity = 2) #une ampoule en plus
        s = pythreejs.Scene(children = [*meshes, lumiere_ambiante, soleil])
        
        c = pythreejs.PerspectiveCamera(aspect=self.width / self.height)
        self.sync(c)
        
        controls = []
        if souris:
            controls.append(pythreejs.OrbitControls(controlling = c))

        click_picker = pythreejs.Picker(controlling=s, event='click')
        controls.append(click_picker)

        return pythreejs.Renderer(width = camera.width, height = camera.height, scene = s, camera = c, controls = controls)

    def lire(self, name = None):
        (t, p, d) = polaires(self.position, self.centre)
        t = degres(t)
        p = degres(p)
        current = dict(
            longitude = t,
            latitude = p,
            distance = d,
            centre_x = self.centre[0],
            centre_y = self.centre[1],
            centre_z = self.centre[2]
        )
        if name:
            return current[name]
        return current

    def ecrire(self, **kwargs):
        current = self.lire()
        for k,v in kwargs.items():
            current[k] = v

        t = radians(current["longitude"])
        p = radians(current["latitude"])
        d = current["distance"]
        cx = current["centre_x"]
        cy = current["centre_y"]
        cz = current["centre_z"]
        x,y,z = cartesiennes(t, p, d, (cx, cy, cz))
        self.position = np.array((x,y,z))
        self.centre = np.array((cx,cy,cz))


    def controles(self, only = None):
        current = self.lire()

        ctrl = []
        w_theta = widgets.FloatSlider(
            value = current['longitude'],
            min = 0, 
            max = 360,
            step = 1,
            description = 'longitude'
        )
        ctrl.append(w_theta)
        w_phi = widgets.FloatSlider(
            value = current['latitude'],
            min = -90, 
            max = 90,
            step = 1,
            description = 'latitude'
        )
        ctrl.append(w_phi)
        w_d = widgets.FloatSlider(
            value = current['distance'],
            min = 0.01, 
            max = 100,
            step = 0.01,
            description = 'distance'
        )
        ctrl.append(w_d)
        w_cx = widgets.FloatSlider(
            value = current["centre_x"],
            min = -10, 
            max = 10,
            step = 0.01,
            description = 'centre_x'
        )
        ctrl.append(w_cx)
        w_cy = widgets.FloatSlider(
            value = current["centre_y"],
            min = -10, 
            max = 10,
            step = 0.01,
            description = 'centre_y'
        )
        ctrl.append(w_cy)
        w_cz = widgets.FloatSlider(
            value = current["centre_z"],
            min = -10, 
            max = 10,
            step = 0.01,
            description = 'centre_z'
        )
        ctrl.append(w_cz)

        def callback(e):
            params = {c.description:c.value for c in ctrl}
            self.ecrire(**params)

        for w in ctrl:
            w.observe(callback)

        filtered = ctrl
        if only:
            filtered = [c for c in ctrl if c.description in only]

        return filtered

def get_picker(renderer):
    for c in renderer.controls:
        if type(c) == pythreejs.controls.Picker_autogen.Picker:
            return c

def scene_pixel(renderer, p):
    P = np.array(renderer.camera.projectionMatrix).reshape((4,4)).T
    WI = np.array(renderer.camera.matrixWorldInverse).reshape((4,4)).T

    hw = 0.5 * renderer.width
    hh = 0.5 * renderer.height

    projected = P@WI@np.array((*p,1))
    projected = projected / projected[3] * np.array((hw,-hh,0,0)) + np.array((hw,hh,0,0))
    return (int(projected[0]),int(projected[1]))

def retrouver(renderer, view = None, controls = None):
    retrouver.counter += 1
    
    side = []
    if view:
        with open(view, "rb") as file:
            img = file.read()
            w_img = widgets.Image(
                value = img, format='png', 
                width=300, height=300, 
                layout = widgets.Layout(
                    border="1px solid black"
                )
            )
            w_img.add_class(f"retrouver_objectif_{retrouver.counter}")
            side.append(w_img)

            download = widgets.HTML("""
                <a href="#" onclick="(() => {{
                    const c = document.querySelector('.retrouver_objectif_{i}') ;
                    let link = document.createElement('a') ;
                    link.href = c.src ;
                    link.download = 'vue.png' ;
                    link.click() ;
                }})()">
                    <button class="p-Widget jupyter-widgets jupyter-button widget-button" width="300px" icon="download">
                    <i class="fa fa-download"></i>image objectif
                    </button>
                </a>
                <a href="#" onclick="(() => {{
                    const c = document.querySelector('.retrouver_{i} canvas') ;
                    let link = document.createElement('a') ;
                    link.href = c.toDataURL() ;
                    link.download = 'image.png' ;
                    link.click() ;
                }})()">
                    <button class="p-Widget jupyter-widgets jupyter-button widget-button" icon="download">
                    <i class="fa fa-download"></i>image actuelle
                    </button>
                </a>
                """.format(file = view, i = retrouver.counter))
            side.append(download)

    
    if controls:
        side.extend(controls)
    
    if renderer:
        renderer.add_class(f"retrouver_{retrouver.counter}")
        picker = get_picker(renderer)
        if picker:
            coords = widgets.Label(
                layout = widgets.Layout(
                    margin = "40px 0 0 0"
                )
            )
            pix = widgets.Label()
            def display_pick(e):
                px = scene_pixel(renderer, picker.point)
                coords.value = f"point 3D : [{picker.point[0]:.2f}, {picker.point[1]:.2f}, {picker.point[2]:.2f}]"
                pix.value = f"pixel : [{px[0]}, {px[1]}]"
            picker.observe(display_pick)
            side.append(coords)
            side.append(pix)
            
        vb = widgets.VBox(side)

        GLbox = widgets.Box(
            [renderer],
            layout = widgets.Layout(
                border = "1px solid black",
                margin = "0 10px 0 0"
            )
        )
        hb = widgets.HBox([GLbox, vb])
        return hb
    else:
        vb = widgets.VBox(side)
        return vb

retrouver.counter = 0
    
def scene_boites():
    boite1 = Box((-2,0,0))
    boite1.push((0,2,2))
    boite1.color = np.array((1.,0.,0.))
    
    boite2 = Box((1.5,0,0.5))
    boite2.push((2.5,4,1.5))
    boite2.color = np.array((0.,1.,0.))
    
    boite3 = Box((1.5,0,-2.5))
    boite3.push((2.5,1,-1.5))
    boite3.color = np.array((0.,0.,1.))
    
    boite4 = Box((-3,-1,-3))
    boite4.push((3,0,3))
    boite4.color = np.array((0.5,0.5,0.5))
    
    return [boite1, boite2, boite3, boite4]
